﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;

using OpenTK;
using OpenTK.Graphics;
using OpenTK.Graphics.OpenGL;

using Meshomatic;

namespace CG1.Ex03.Meshes
{
    class MeshBatch
    {
        #region Properties

        public MeshData m;
        uint dataBuffer;
        uint indexBuffer;
        uint tex;
        Int32 ShaderHandle;
        int vertOffset, normOffset, texcoordOffset;

        #endregion

        #region Init

        /// <summary>
        /// Set the MeshBatch 
        /// </summary>
        public MeshBatch(String ObjFile, String TexFile)
        {
            m = new ObjLoader().LoadFile(ObjFile);
            tex = LoadTex(TexFile);

            GL.Enable(EnableCap.DepthTest);
            GL.Enable(EnableCap.Texture2D);
            GL.EnableClientState(ArrayCap.VertexArray);
            GL.EnableClientState(ArrayCap.NormalArray);
            GL.EnableClientState(ArrayCap.TextureCoordArray);

            ShaderHandle = CompileShaders();
            GL.UseProgram(ShaderHandle);
            LoadBuffers(m);
        }

        #endregion

        #region Shaders

        string vShaderSource = @"
        void main() {
	        gl_Position = ftransform();
	        gl_TexCoord[0] = gl_MultiTexCoord0;
        }
        ";
        string fShaderSource = @"
        uniform sampler2D tex;
        void main() {
	        gl_FragColor = texture2D(tex, gl_TexCoord[0].st);
        }
        ";

        /// <summary>
        /// Create handles and compile vertex and fragment shaders
        /// </summary>
        int CompileShaders()
        {
            int programHandle, vHandle, fHandle;
            vHandle = GL.CreateShader(ShaderType.VertexShader);
            GL.ShaderSource(vHandle, vShaderSource);
            GL.CompileShader(vHandle);

            fHandle = GL.CreateShader(ShaderType.FragmentShader);
            GL.ShaderSource(fHandle, fShaderSource);
            GL.CompileShader(fHandle);

            programHandle = GL.CreateProgram();
            GL.AttachShader(programHandle, vHandle);
            GL.AttachShader(programHandle, fHandle);
            GL.LinkProgram(programHandle);
            return programHandle;
        }

        #endregion

        #region Buffers

        /// <summary>
        /// Set everything to work with buffers. And then we will just display buffers each time.
        /// </summary>
        public void LoadBuffers(MeshData m)
        {
            
            float[] verts, norms, texcoords;
            uint[] indices;
            m.OpenGLArrays(out verts, out norms, out texcoords, out indices);
            GL.GenBuffers(1, out dataBuffer);
            GL.GenBuffers(1, out indexBuffer);

            //Info: Set up data for VBO.
            //      We're going to use one VBO for all geometry, and stick it in (VVVVNNNNCCCC) order. Non interleaved.
            int buffersize = (verts.Length + norms.Length + texcoords.Length);
            float[] bufferdata = new float[buffersize];
            vertOffset = 0;
            normOffset = verts.Length;
            texcoordOffset = (verts.Length + norms.Length);

            verts.CopyTo(bufferdata, vertOffset);
            norms.CopyTo(bufferdata, normOffset);
            texcoords.CopyTo(bufferdata, texcoordOffset);

            bool v = false;
            for (int i = texcoordOffset; i < bufferdata.Length; i++)
            {
                if (v)
                {
                    bufferdata[i] = 1 - bufferdata[i];
                    v = false;
                }
                else
                {
                    v = true;
                }
            }

            //Info: Load geometry data
            GL.BindBuffer(BufferTarget.ArrayBuffer, dataBuffer);
            GL.BufferData<float>(BufferTarget.ArrayBuffer, (IntPtr)(buffersize * sizeof(float)), bufferdata,
                          BufferUsageHint.StaticDraw);

            //Info: Load index data
            GL.BindBuffer(BufferTarget.ElementArrayBuffer, indexBuffer);
            GL.BufferData<uint>(BufferTarget.ElementArrayBuffer,
                          (IntPtr)(indices.Length * sizeof(uint)), indices, BufferUsageHint.StaticDraw);
        }

        /// <summary>
        /// Display buffers. There are our textures, vertices etc. At the end, render object with draw elements.
        /// </summary>
        public void DrawBuffer()
        {
            GL.UseProgram(ShaderHandle);
            //Info: Push current Array Buffer state so we can restore it later
            GL.PushClientAttrib(ClientAttribMask.ClientVertexArrayBit);

            GL.ClientActiveTexture(TextureUnit.Texture0);
            GL.BindTexture(TextureTarget.Texture2D, tex);

            GL.BindBuffer(BufferTarget.ArrayBuffer, dataBuffer);
            //Info: Normal buffer
            GL.NormalPointer(NormalPointerType.Float, 0, (IntPtr)(normOffset * sizeof(float)));

            //Info: TexCoord buffer
            GL.TexCoordPointer(2, TexCoordPointerType.Float, 0, (IntPtr)(texcoordOffset * sizeof(float)));

            //Info: Vertex buffer
            GL.VertexPointer(3, VertexPointerType.Float, 0, (IntPtr)(vertOffset * sizeof(float)));

            //Info: Index array
            GL.BindBuffer(BufferTarget.ElementArrayBuffer, indexBuffer);
            GL.DrawElements(BeginMode.Triangles, m.Tris.Length * 3, DrawElementsType.UnsignedInt, IntPtr.Zero);

            //Info: Restore the state
            GL.PopClientAttrib();
        }

        #endregion

        #region Load Texture

        /// <summary>
        /// Load texture as bitmap with power of 2. Also set filters and other properties of texture in OpenGL.
        /// </summary>
        static uint LoadTex(string file)
        {
            Bitmap bitmap = new Bitmap(file);
            //Info: Warning: Texture must be a power of 2!
            uint texture;
            GL.Hint(HintTarget.PerspectiveCorrectionHint, HintMode.Nicest);

            GL.GenTextures(1, out texture);
            GL.BindTexture(TextureTarget.Texture2D, texture);

            BitmapData data = bitmap.LockBits(new System.Drawing.Rectangle(0, 0, bitmap.Width, bitmap.Height),
                ImageLockMode.ReadOnly, System.Drawing.Imaging.PixelFormat.Format32bppArgb);

            GL.TexImage2D(TextureTarget.Texture2D, 0, PixelInternalFormat.Rgba, data.Width, data.Height, 0,
                OpenTK.Graphics.OpenGL.PixelFormat.Bgra, PixelType.UnsignedByte, data.Scan0);
            bitmap.UnlockBits(data);

            GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMinFilter, (int)TextureMinFilter.Linear);
            GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMagFilter, (int)TextureMagFilter.Linear);

            return texture;
        }

        #endregion
    }
}
